<?php

namespace Drupal\radioactivity;

use Drupal\Component\Serialization\SerializationInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Database\DatabaseException;

/**
 * Defines a default incident storage.
 */
class DefaultIncidentStorage implements IncidentStorageInterface {

  /**
   * The table name.
   */
  const TABLE_NAME = 'radioactivity_incident';

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $connection;

  /**
   * The serialization class to use.
   *
   * @var \Drupal\Component\Serialization\SerializationInterface
   */
  protected $serializer;

  /**
   * DefaultIncidentStorage constructor.
   *
   * @param \Drupal\Core\Database\Connection $connection
   *   The database connection.
   * @param \Drupal\Component\Serialization\SerializationInterface $serializer
   *   The serialization class to use.
   */
  public function __construct(Connection $connection, SerializationInterface $serializer) {
    $this->connection = $connection;
    $this->serializer = $serializer;
  }

  /**
   * {@inheritdoc}
   */
  public function addIncident(IncidentInterface $incident) {
    try {
      $this->doAdd($incident);
    }
    catch (\Exception $exception) {
      $this->ensureTableExists();
      $this->doAdd($incident);
    }
  }

  /**
   * Adds incident data to the database.
   *
   * @param \Drupal\radioactivity\IncidentInterface $incident
   *   The incident object.
   */
  protected function doAdd(IncidentInterface $incident) {
    $this->connection->insert(self::TABLE_NAME)
      ->fields([
        'incident' => $this->serializer->encode($incident),
        'entity_type' => $incident->getEntityTypeId(),
      ])
      ->execute();
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Exception
   *   Throws exception when an unexpected database error occurs.
   */
  public function getIncidents(): array {
    try {
      $result = $this->connection->select(self::TABLE_NAME, 'ri')
        ->fields('ri', ['iid', 'incident'])
        ->orderBy('ri.iid', 'ASC')
        ->execute();

      $values = [];
      foreach ($result as $item) {
        if ($item) {
          $values[] = $this->serializer->decode($item->incident);
        }
      }
      return $values;
    }
    catch (\Exception $exception) {
      if ($this->connection->schema()->tableExists(self::TABLE_NAME)) {
        throw $exception;
      }

      $this->ensureTableExists();
      return [];
    }
  }

  /**
   * {@inheritdoc}
   *
   * @throws \Exception
   *   Throws exception when an unexpected database error occurs.
   */
  public function getIncidentsByType(string $entity_type = ''): array {
    try {
      $query = $this->connection->select(self::TABLE_NAME, 'ri');
      $query->fields('ri', ['iid', 'incident', 'entity_type']);
      $query->orderBy('ri.iid', 'ASC');
      if ($entity_type) {
        $query->condition('entity_type', $entity_type);
      }
      $result = $query->execute();

      $incidents = [];
      foreach ($result as $item) {
        /** @var \Drupal\radioactivity\IncidentInterface $incident */
        $incident = $this->serializer->decode($item->incident);
        $incidents[$incident->getEntityTypeId()][$incident->getEntityId()][] = $incident;
      }

      return $incidents ?: [[]];
    }
    catch (\Exception $exception) {
      if ($this->connection->schema()->tableExists(self::TABLE_NAME)) {
        throw $exception;
      }

      $this->ensureTableExists();
      return [[]];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function clearIncidents() {
    try {
      $this->connection
        ->truncate(self::TABLE_NAME)
        ->execute();
    }
    catch (\Exception $exception) {
      $this->ensureTableExists();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function injectSettings(array &$page) {
    global $base_url;
    $page['#attached']['drupalSettings']['radioactivity']['type'] = 'default';
    $page['#attached']['drupalSettings']['radioactivity']['endpoint'] = $base_url . '/radioactivity/emit';
  }

  /**
   * Check if the table exists and create it if not.
   *
   * Thanks to \Drupal\Core\Batch\BatchStorage::ensureTableExists.
   *
   * @return bool
   *   Returns true if the database table has been created.
   */
  protected function ensureTableExists() {
    try {
      $database_schema = $this->connection->schema();
      if (!$database_schema->tableExists(static::TABLE_NAME)) {
        $schema_definition = $this->schemaDefinition();
        $database_schema->createTable(static::TABLE_NAME, $schema_definition);
        return TRUE;
      }
    }
      // If another process has already created the table, attempting to
      // recreate it will throw an exception. In this case just catch the
      // exception and do nothing.
    catch (DatabaseException $e) {
      return TRUE;
    }
    return FALSE;
  }

  /**
   * Defines the schema for the batch table.
   *
   * @internal
   */
  public function schemaDefinition() {
    return [
      'description' => 'Stores details about batches (processes that run in multiple HTTP requests).',
      'fields' => [
        'iid' => [
          'description' => 'Unique incident ID.',
          'type' => 'serial',
          'not null' => TRUE,
        ],
        'entity_type' => [
          'description' => 'The entity type this incident belongs to.',
          'type' => 'varchar_ascii',
          'length' => 64,
          'not null' => TRUE,
        ],
        'incident' => [
          'description' => 'The incident object.',
          'type' => 'blob',
          'not null' => FALSE,
          'size' => 'normal',
        ],
      ],
      'primary key' => ['iid'],
      'indexes' => [
        'entity_type' => ['entity_type'],
      ],
    ];
  }

}
